

#include <unistd.h>
#include <errno.h>
#include <sys/types.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "common.h"
#include "message.h"
#include "buffer.h"
#include "transfer.h"

#include "merge.h"
#include "stats.h"

extern int NumRunningLinks;

extern STATUS_TABLE *status_table;
extern BUFFER_TABLE *buffer_table;
extern LINK_TABLE *link_table[NUMlinks];

extern int * MERGE_RunOptions;

extern int * StatsMem[NUMCOUNTERS];

extern int do_storage (char *, int);

extern  int BufferSize;

extern UINT64 lastTime;      /*  time of last data item  */
extern UINT64 oldestTime;    /*  from last data item  */
extern int oldestLink;       /*  from last data item */
extern UINT wr4863;        /*  from last Info#5 */

extern ItDesc_Ptr OutputBufferPtr;
extern int OutputBufferLen;
extern char * OutputBuffer;

extern int dataStart[NUMlinks];
extern int bufferNumber[NUMlinks];
extern char * dataBuffer[NUMlinks];
extern int dataLen[NUMlinks];


   int do_merge () {

/*
    processes link buffer qs & generates merged data until 1 link q becomes empty
    then returns with that q id
*/

       int rc =-1;
       int done = FALSE;
       int ID, i, j;
       int module, icode;
       NewItDesc_Ptr ps, os;

       sprintf(message_buffer, "do_merge merging %d links  0x%016llx", NumRunningLinks, lastTime);
       report_message(MSG_TRACE+0);

       while (done == FALSE) {
          os = NULL;
          oldestTime = 0;

         for (ID = 0; ID < status_table->num_links; ID++) {
             if (link_table[ID]->link_state != LINK_GOING) continue;

             ps = (NewItDesc_Ptr)dataBuffer[ID];

             sprintf(message_buffer, "merge#1 %d 0x%x 0x%016llx 0x%016llx %d", ID, ps, ps->Time, oldestTime, dataStart[ID]);             
             report_message(MSG_DEBUG+0);

             ps += dataStart[ID];

             if (ps->Time < oldestTime || oldestTime == 0) {
 
                sprintf(message_buffer, "merge#2 %d %d 0x%016llx 0x%016llx %d ",i, ID, ps->Time, oldestTime, oldestLink);
                report_message(MSG_DEBUG+0);

                os = ps;
                oldestTime = os->Time;
                oldestLink = ID;

             }
          }     ;//  (ID = 0; ID < status_table->num_links; ID++)  - oldest TS now found

          sprintf(message_buffer, "merge#3 oldest %d 0x%8x 0x%016llx %d %d",oldestLink,os,oldestTime,dataStart[oldestLink],dataLen[oldestLink]);
          report_message(MSG_DEBUG+0);

          if (os != NULL) {   // be cautious - be sure an oldest has really been found

            dataStart[oldestLink]++;

            if (dataStart[oldestLink] >= dataLen[oldestLink]) {
               dataStart[oldestLink] = 0;

               sprintf(message_buffer, "merge_server (merge done=TRUE) %d", oldestLink);
               report_message(MSG_DEBUG+0);
 
               i = bufferNumber[oldestLink];
               j = oldestLink*N_DATA_BUFFERS;

               buffer_table->buffer_status[i+j] = DONE;            //  flag all data in this buffer has been processed
               rc = oldestLink;
               done = TRUE;
            }

            if (os->Time < lastTime) {

              sprintf(message_buffer, "before previous time 0x%016llx < 0x%016llx", os->Time, lastTime);
              report_message(MSG_WARNING+0);

              lastTime = os->Time;
              (*StatsMem[MTIMEERR])++;   

            }  else {

// For now just output the data in the standard format

            sprintf(message_buffer, "merge_server (OP) %d 0x%016llx %lx %lx %d", oldestLink, os->Time, os->Data, os->Timestamp, OutputBufferLen);
            report_message(MSG_TRACE+0);

            lastTime = os->Time;
            (*StatsMem[MTIMEOK])++;

            module = (os->Data >> 24) & 63;
            icode = (os->Data >> 20) & 15;

            if (icode == WR_CODE) {wr4863 = (os->Data & 0x0000ffff); goto L00;}

            if (icode == SYNC_CODE) {  /* Info #4   needs to be coupled with Info #5 */
                
             wr4863 = 0;

                OutputBufferPtr->data = (2 << 30) | (module << 24) | (WR_CODE << 20) | wr4863;  /* generate the Info #5 */
                OutputBufferPtr->timestamp = os->Timestamp;
                OutputBufferPtr++;
                OutputBufferLen += sizeof(ItDesc);
            }

            OutputBufferPtr->data = os->Data;
            OutputBufferPtr->timestamp = os->Timestamp;
            OutputBufferPtr++;
            OutputBufferLen += sizeof(ItDesc);

            if (OutputBufferLen >= (BufferSize - 32 - sizeof(ItDesc) - sizeof(ItDesc))) { /* allow for info 5 + info 4 */

// No space for another item - the output buffer is full

               if ((*MERGE_RunOptions & 2) != 0) {rc = do_storage (OutputBuffer, OutputBufferLen);}
               OutputBufferPtr = (ItDesc_Ptr)OutputBuffer;
               OutputBufferLen = 0;

               (*StatsMem[MBUFFERS])++;           /* counter for number of merged data buffers */

            }
L00:       continue;
           }    ;//  end of os->Time < lastTime

         }    ;//  end of if (os)

      } // end of     while (done == FALSE)    loop - one of the input buffers is empty

      return rc;
   }




